// 2 sensor model of the FiGo algorithm. This version contains synchronisation of clock updates, and gossiping.
//
// Matt Webster, University of Liverpool, matt@liverpool.ac.uk. April 2018.
// Works with PRISM 4.3.1.

dtmc


// Sensor states
const int transmit			= 0;
const int clockCycleCheck		= 1;
const int listen			= 2;
const int sync1				= 3;
const int sync2				= 4;
const int sync3				= 5;
const int sync4				= 6;
const int updateClock			= 7;

// other constants
const int cycleLength 			= 20;
const int refractoryPeriod		= floor(cycleLength/2);  	// i.e., cycleLength/2
const int dutyCycle			= cycleLength;
const int sameThreshold			= 2; 
const int metadataMax			= 3;
const double pUpdateMetadata		= 0.5;


// inboxes for s1, s2
// new values sent to these sensors are stored here.
global s1InboxClock: [0..cycleLength] init 0;
global s1InboxMeta:  [0..metadataMax] init 0;
global s1InboxFull: bool init false;
global s2InboxClock: [0..cycleLength] init 0;
global s2InboxMeta:  [0..metadataMax] init 0;
global s2InboxFull: bool init false;

const int s1NextBroadcast			= 10;  	// should be randomly set
const int s1ClockInit				= 5;

module sensor1

	s1State : [0..9] init transmit;
	s1Clock : [0..cycleLength] init s1ClockInit;
	s1LocalMetadata : [0..metadataMax] init 0;
	s1SameCount : [0..2] init 0;


	// if local clock == next broadcast and same count < same threshold then
	// transmit local clock and local metadata value
	// restart at top of loop
	[] s1State=transmit & (s1Clock=s1NextBroadcast | s1Clock=(mod(s1NextBroadcast+refractoryPeriod,cycleLength))) & s1SameCount<sameThreshold -> (s2InboxClock'=s1Clock) & (s2InboxMeta'=s1LocalMetadata) & (s2InboxFull'=true) & (s1SameCount'=0) & (s1State'=updateClock);
	[] s1State=transmit & !((s1Clock=s1NextBroadcast | s1Clock=(mod(s1NextBroadcast+refractoryPeriod,cycleLength))) & s1SameCount<sameThreshold) -> (s1State'=clockCycleCheck);

	// Algorithm: if clock ≤ cycle length then listen
	[] s1State=clockCycleCheck & s1Clock<cycleLength -> (s1State'=listen);
	[] s1State=clockCycleCheck & !(s1Clock<cycleLength) -> (s1State'=updateClock);
 
	// Algorithm: if A message is overheard
	// so, if the inbox clock has changed from last time, then a message must have been received, so updated the clock and the metadata and go into sync1 state
	[] s1State=listen & s1InboxFull -> (s1InboxFull'=false) & (s1State'=sync1);
	// if there is no message, then jump to the update clock state which advances the clock before going back to the start of the loop
	[] s1State=listen & !s1InboxFull -> (s1State'=updateClock);

	// if local clock > refractory period then
	// adjust local clock to average of local clock and time in message
	[] s1State=sync1 & s1Clock>=refractoryPeriod & diff<=floor(cycleLength/2) -> (s1Clock'=s1avg1) & (s1State'=sync2);
	[] s1State=sync1 & s1Clock>=refractoryPeriod & diff>floor(cycleLength/2) -> (s1Clock'=s1avg2) & (s1State'=sync2);
	[] s1State=sync1 & !( s1Clock>=refractoryPeriod ) -> (s1State'=sync2); 
	// if A message is overheard AND local clock > refractory period then
	// adjust local clock to average of local clock and time in message
	// end if
	[] s1State=sync2 & s1InboxMeta>s1LocalMetadata -> (s1LocalMetadata'=s1InboxMeta) & (s1State'=updateClock); 
	[] s1State=sync2 & !(s1InboxMeta>s1LocalMetadata) -> (s1State'=sync3); 
	
	// else if metadata < local metadata AND same count < 1 then transmit data now
	[] s1State=sync3 & s1InboxMeta<s1LocalMetadata & s1SameCount<1 -> (s2InboxClock'=s1Clock) & (s2InboxMeta'=s1LocalMetadata) & (s2InboxFull'=true) & (s1State'=updateClock);
	[] s1State=sync3 & !(s1InboxMeta<s1LocalMetadata & s1SameCount<1) -> (s1State'=sync4);
	// else if the message contains metadata == local metadata AND time in message ==localclock then same count = same count + 1
	[] s1State=sync4 & s1InboxMeta=s1LocalMetadata & s1InboxClock=s1Clock & (s1SameCount<1) -> (s1SameCount'=s1SameCount+1) & (s1State'=updateClock);	
	[] s1State=sync4 & s1InboxMeta=s1LocalMetadata & s1InboxClock=s1Clock & (s1SameCount>=1) -> (s1State'=updateClock);	
	[] s1State=sync4 & !(s1InboxMeta=s1LocalMetadata & s1InboxClock=s1Clock) -> (s1State'=updateClock);
	// local clock = local clock + 1
	[tick] s1State=updateClock & s1Clock<cycleLength -> (s1Clock'=s1Clock+1) & (s1State'=transmit);
	[tick] s1State=updateClock & s1Clock=cycleLength & s1LocalMetadata<metadataMax -> (1-pUpdateMetadata): (s1Clock'=0) & (s1SameCount'=0) & (s1State'=transmit)
									   + 	pUpdateMetadata: (s1Clock'=0) & (s1SameCount'=0) & 
										(s1LocalMetadata'=s1LocalMetadata+1) & (s1State'=transmit);
	[tick] s1State=updateClock & s1Clock=cycleLength & s1LocalMetadata>=metadataMax -> (s1Clock'=0) & (s1SameCount'=0) & (s1State'=transmit);

endmodule

const int s2NextBroadcast			= 3;  	// should be randomly set


const int s2ClockInit  = 0;

module sensor2


	s2State : [0..9] init transmit;
	s2Clock : [0..cycleLength] init s2ClockInit;
	s2LocalMetadata : [0..metadataMax] init 1;
	s2SameCount : [0..2] init 0;


	// if local clock == next broadcast and same count < same threshold then
	// transmit local clock and local metadata value
	// restart at top of loop
	[] s2State=transmit & (s2Clock=s2NextBroadcast | s2Clock=(mod(s2NextBroadcast+refractoryPeriod,cycleLength))) & s2SameCount<sameThreshold -> (s1InboxClock'=s2Clock) & (s1InboxMeta'=s2LocalMetadata) & (s1InboxFull'=true) & (s2SameCount'=0) & (s2State'=updateClock);
	[] s2State=transmit & !((s2Clock=s2NextBroadcast | s2Clock=(mod(s2NextBroadcast+refractoryPeriod,cycleLength))) & s2SameCount<sameThreshold) -> (s2State'=clockCycleCheck);

	
	// Algorithm: if clock ≤ cycle length then listen
	[] s2State=clockCycleCheck & s2Clock<cycleLength -> (s2State'=listen);
	[] s2State=clockCycleCheck & !(s2Clock<cycleLength) -> (s2State'=updateClock);

	// Algorithm: if A message is overheard
	// so, if the inbox clock has changed from last time, then a message must have been received, so updated the clock and the metadata and go into sync1 state
	[] s2State=listen & s2InboxFull -> (s2InboxFull'=false) & (s2State'=sync1);
	// if there is no message, then jump to the update clock state which advances the clock before going back to the start of the loop
	[] s2State=listen & !s2InboxFull -> (s2State'=updateClock);

	// if local clock > refractory period then 
	// adjust local clock to average of local clock and time in message
	[] s2State=sync1 & s2Clock>=refractoryPeriod & diff<=floor(cycleLength/2) -> (s2Clock'=s2avg1) & (s2State'=sync2);
	[] s2State=sync1 & s2Clock>=refractoryPeriod & diff>floor(cycleLength/2) -> (s2Clock'=s2avg2) & (s2State'=sync2);
	[] s2State=sync1 & !( s2Clock>=refractoryPeriod ) -> (s2State'=sync2);
	// if A message is overheard AND local clock > refractory period then
	// adjust local clock to average of local clock and time in message
	// end if
	// if metadata > local metadata then local metadata := metadata
	[] s2State=sync2 & s2InboxMeta>s2LocalMetadata -> (s2LocalMetadata'=s2InboxMeta) & (s2State'=updateClock); 
	[] s2State=sync2 & !(s2InboxMeta>s2LocalMetadata) -> (s2State'=sync3); 
	// else if metadata < local metadata AND same count < 1 then transmit data now
	[] s2State=sync3 & s2InboxMeta<s2LocalMetadata & s2SameCount<1 -> (s1InboxClock'=s2Clock) & (s1InboxMeta'=s2LocalMetadata) & (s1InboxFull'=true) & (s2State'=updateClock);
	[] s2State=sync3 & !(s2InboxMeta<s2LocalMetadata & s2SameCount<1) -> (s2State'=sync4);
	// else if the message contains metadata == local metadata AND time in message ==localclock then same count = same count + 1
	[] s2State=sync4 & s2InboxMeta=s2LocalMetadata & s2InboxClock=s2Clock & (s2SameCount<1) -> (s2SameCount'=s2SameCount+1) & (s2State'=updateClock);	
	[] s2State=sync4 & s2InboxMeta=s2LocalMetadata & s2InboxClock=s2Clock & (s2SameCount>=1) -> (s2State'=updateClock);	
	[] s2State=sync4 & !(s2InboxMeta=s2LocalMetadata & s2InboxClock=s2Clock) -> (s2State'=updateClock);
	// local clock = local clock + 1
	[tick] s2State=updateClock & s2Clock=cycleLength -> (s2Clock'=0) & (s2SameCount'=0) & (s2State'=transmit);
	[tick] s2State=updateClock & s2Clock<cycleLength -> (s2Clock'=s2Clock+1) & (s2State'=transmit);

endmodule




formula diff = max(s1Clock,s2Clock)-min(s1Clock,s2Clock);

// averages for s1
formula s1avg1 = ceil((s1Clock+s1InboxClock)/2);
formula s1avg2 = mod(ceil((s1Clock+s1InboxClock+cycleLength)/2),cycleLength);
// averages for s2
formula s2avg1 = ceil((s2Clock+s2InboxClock)/2);
formula s2avg2 = mod(ceil((s2Clock+s2InboxClock+cycleLength)/2),cycleLength);



// END OF FILE
